from tempfile import NamedTemporaryFile

import keras
import pytest
from keras import ops

from mingru_keras import MinGRU
from mingru_keras.core import Blellochs_method, sequential_method


@pytest.mark.parametrize("b,n,d", [(32, 10, 8), (1, 1000, 1)])
def test_Blellochs_method(b, n, d):
    X = keras.random.normal((b, n, d))
    Z = keras.random.uniform((b, n, d))
    H_desired = sequential_method(X, Z)
    H_actual = Blellochs_method(X, Z)

    assert ops.max(ops.abs(H_actual - H_desired)) < 1e-6


@pytest.mark.parametrize("b,n,i,d", [(16, 10, 8, 16), (32, 100, 2, 64)])
def test_MinGRU(b, n, i, d):
    layer = MinGRU(d)
    X = keras.random.normal((b, n, i))
    Y = layer(X)
    assert Y.shape == (b, n, d)

    # Count parameters.
    assert layer.gate.count_params() == i * d + d
    assert layer.candidate.count_params() == i * d + d


@pytest.mark.parametrize("b,n,i,d", [(32, 100, 8, 16)])
def test_saving(b, n, i, d):
    model = keras.Sequential([MinGRU(d)])
    model.build((None, None, i))
    model.summary()

    with NamedTemporaryFile("wb", suffix=".keras") as f:
        keras.saving.save_model(model, f.name)
        model2 = keras.saving.load_model(f.name)
        pass

    X = keras.random.normal((b, n, i))
    Y1 = model(X)
    Y2 = model2(X)

    assert ops.max(ops.abs(Y1 - Y2)) < 1e-4



import keras
from keras import ops


def sequential_method(H_tilde, Z):
    _, n, _ = H_tilde.shape
    h = ops.zeros_like(H_tilde[:, 0])
    H = []
    for i in range(n):
        h = h + Z[:, i, :] * (H_tilde[:, i, :] - h)
        H.append(h)
    return ops.stack(H, axis=1)


def Blelloch_operator(prev, curr):
    prev_keep, prev_hidden = prev
    curr_keep, curr_hidden = curr
    keep = prev_keep * curr_keep
    hidden = prev_hidden * curr_keep + curr_hidden
    return keep, hidden


def Blellochs_method(H_tilde, Z, axis=-2):
    _, H = ops.associative_scan(Blelloch_operator, ((1 - Z), Z * H_tilde), axis=axis)
    return H


class MinGRU(keras.Layer):
    def __init__(self, units, **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.gate = keras.layers.Dense(units, activation="sigmoid")
        self.candidate = keras.layers.Dense(units)

    def build(self, input_shape):
        super().build(input_shape)
        self.gate.build(input_shape)
        self.candidate.build(input_shape)

    def compute_output_shape(self, input_shape):
        b, t, _ = input_shape
        return b, t, self.units

    def call(self, X):
        Z = self.gate(X)
        H_tilde = self.candidate(X)
        H = Blellochs_method(H_tilde, Z)

        return H

    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units})
        return config
